import hcp
from hcp.tests import config as tconf


def test_set_eog_channels():
    pass


def test_apply_ica():
    pass


def test_apply_ref_correction():
    pass


def test_map_ch_coords_to_mne():
    pass


def test_interpolate_missing():
    pass
